from collections.abc import Callable
from typing import Any

from lomond import WebSocket
from lomond.persist import persist
from pydantic import ValidationError

from ..types.clob_types import ApiCreds
from ..types.websockets_types import (
    CommentEvent,
    LiveDataOrderMatchEvent,
    LiveDataTradeEvent,
    OrderBookSummaryEvent,
    OrderEvent,
    PriceChangeEvent,
    QuoteEvent,
    ReactionEvent,
    RequestEvent,
    TickSizeChangeEvent,
    TradeEvent,
)


def _process_market_event(event):
    try:
        event = event.json
        for message in event:
            match message["event_type"]:
                case "book":
                    print(OrderBookSummaryEvent(**message), "\n")
                case "price_change":
                    print(PriceChangeEvent(**message), "\n")
                case "tick_size_change":
                    print(TickSizeChangeEvent(**message), "\n")
    except ValidationError as e:
        print(event.text)
        print(e.errors(), "\n")

def _process_user_event(event):
    try:
        event = event.json
        for message in event:
            match message["event_type"]:
                case "order":
                    print(OrderEvent(**message), "\n")
                case "trade":
                    print(TradeEvent(**message), "\n")
    except ValidationError as e:
        print(event.text)
        print(e.errors(), "\n")

def _process_live_data_event(event):
    try:
        message = event.json
        match message["type"]:
            case "trades":
                print(LiveDataTradeEvent(**message), "\n")
            case "orders_matched":
                print(LiveDataOrderMatchEvent(**message), "\n")
            case "comment_created" | "comment_removed":
                print(CommentEvent(**message), "\n")
            case "reaction_created" | "reaction_removed":
                print(ReactionEvent(**message), "\n")
            case "request_created" | "request_edited" | "request_canceled" | "request_expired":
                print(RequestEvent(**message), "\n")
            case "quote_created" | "quote_edited" | "quote_canceled" | "quote_expired":
                print(QuoteEvent(**message), "\n")
    except ValidationError as e:
        print(event.text)
        print(e.errors(), "\n")

class PolymarketWebsocketsClient:
    def __init__(self):
        self.url_market = "wss://ws-subscriptions-clob.polymarket.com/ws/market"
        self.url_user = "wss://ws-subscriptions-clob.polymarket.com/ws/user"
        self.url_live_data = "wss://ws-live-data.polymarket.com"

    def market_socket(self, token_ids: list[str], process_event: Callable = _process_market_event):
        """
        Connect to the market websocket and subscribe to market events for specific token IDs.

        Args:
            token_ids: List of token IDs to subscribe to
            process_event: Callback function to process received events

        """
        websocket = WebSocket(self.url_market)
        for event in persist(websocket):  # persist automatically reconnects
            if event.name == "ready":
                websocket.send_json(
                    assets_ids=token_ids,
                )
            elif event.name == "text":
                process_event(event)

    def user_socket(self, creds: ApiCreds, process_event: Callable = _process_user_event):
        """
        Connect to the user websocket and subscribe to user events.

        Args:
            creds: API credentials for authentication
            process_event: Callback function to process received events

        """
        websocket = WebSocket(self.url_user)
        for event in persist(websocket):
            if event.name == "ready":
                websocket.send_json(
                    auth = creds.model_dump(by_alias=True),
                )
            elif event.name == "text":
                process_event(event)

    def live_data_socket(self, subscriptions: list[dict[str, Any]], process_event: Callable = _process_live_data_event):
        """
        Connect to the live data websocket and subscribe to specified events.

        Args:
            subscriptions: List of subscription configurations
            process_event: Callback function to process received events

        """
        websocket = WebSocket(self.url_live_data)
        for event in persist(websocket):
            if event.name == "ready":
                websocket.send_json(
                    action="subscribe",
                    subscriptions=subscriptions,
                )
            elif event.name == "text":
                process_event(event)



import hashlib
from decimal import Decimal
from math import ceil, floor

from ...types.clob_types import OrderBookSummary, TickSize


def round_down(x: float, sig_digits: int) -> float:
    return floor(x * (10**sig_digits)) / (10**sig_digits)


def round_normal(x: float, sig_digits: int) -> float:
    return round(x * (10**sig_digits)) / (10**sig_digits)


def round_up(x: float, sig_digits: int) -> float:
    return ceil(x * (10**sig_digits)) / (10**sig_digits)


def to_token_decimals(x: float) -> int:
    f = (10**6) * x
    if decimal_places(f) > 0:
        f = round_normal(f, 0)
    return int(f)


def decimal_places(x: float) -> int:
    """
    Returns the number of decimal places in a float.

    Assumes x is always a finite, non-special value (not NaN or Infinity).
    """
    exponent = Decimal(str(x)).as_tuple().exponent
    if not isinstance(exponent, int):
        msg = "Input must be a finite float."
        raise TypeError(msg)
    return abs(exponent)



def generate_orderbook_summary_hash(orderbook: OrderBookSummary) -> str:
    """Compute hash while forcing empty string for hash field."""
    server_hash = orderbook.hash
    orderbook.hash = ""
    computed_hash = hashlib.sha1(
        str(orderbook.model_dump_json(by_alias=True)).encode("utf-8"),
    ).hexdigest()
    orderbook.hash = server_hash
    return computed_hash


def order_to_json(order, owner, order_type) -> dict:
    return {"order": order.dict(), "owner": owner, "orderType": order_type.value}


def is_tick_size_smaller(a: TickSize, b: TickSize) -> bool:
    return float(a) < float(b)


def price_valid(price: float, tick_size: TickSize) -> bool:
    return float(tick_size) <= price <= 1 - float(tick_size)



import json
from datetime import datetime
from typing import Optional, Union
from urllib.parse import urljoin

import httpx

from ..types.gamma_types import Event, GammaMarket, QueryEvent


class PolymarketGammaClient:
    def __init__(self, base_url: str = "https://gamma-api.polymarket.com"):
        self.base_url = base_url
        self.client = httpx.Client(http2=True, timeout=30.0)

    def _build_url(self, endpoint: str) -> str:
        return urljoin(self.base_url, endpoint)

    def get_markets(
            self,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            order: Optional[str] = None,
            ascending: bool = True,
            ids: Optional[list[int]] = None,
            slugs: Optional[list[str]] = None,
            archived: Optional[bool] = None,
            active: Optional[bool] = None,
            closed: Optional[bool] = None,
            token_ids: Optional[list[str]] = None,
            condition_ids: Optional[list[str]] = None,
            liquidity_num_min: Optional[float] = None,
            liquidity_num_max: Optional[float] = None,
            volume_num_min: Optional[float] = None,
            volume_num_max: Optional[float] = None,
            start_date_min: Optional[datetime] = None,
            start_date_max: Optional[datetime] = None,
            end_date_min: Optional[datetime] = None,
            end_date_max: Optional[datetime] = None,
            tag_id: Optional[int] = None,
            related_tags: Optional[bool] = False,
    ) -> list[GammaMarket]:
        params = {}
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        if order:
            params["order"] = order
            params["ascending"] = ascending
        if ids:
            params["id"] = ids
        if slugs:
            params["slug"] = slugs
        if archived is not None:
            params["archived"] = archived
        if active is not None:
            params["active"] = active
        if closed is not None:
            params["closed"] = closed
        if token_ids:
            params["clob_token_ids"] = token_ids
        if condition_ids:
            params["condition_ids"] = condition_ids
        if liquidity_num_min:
            params["liquidity_num_min"] = liquidity_num_min
        if liquidity_num_max:
            params["liquidity_num_max"] = liquidity_num_max
        if volume_num_min:
            params["volume_num_min"] = volume_num_min
        if volume_num_max:
            params["volume_num_max"] = volume_num_max
        if start_date_min:
            params["start_date_min"] = start_date_min.isoformat()
        if start_date_max:
            params["start_date_max"] = start_date_max.isoformat()
        if end_date_min:
            params["end_date_min"] = end_date_min.isoformat()
        if end_date_max:
            params["end_date_max"] = end_date_max.isoformat()
        if tag_id:
            params["tag_id"] = tag_id
            if related_tags:
                params["related_tags"] = related_tags

        response = self.client.get(self._build_url("/markets"), params=params)
        response.raise_for_status()
        return [GammaMarket(**market) for market in response.json()]

    def get_market(self, market_id: str) -> GammaMarket:
        """Get a GammaMarket by market_id."""
        response = self.client.get(self._build_url(f"/markets/{market_id}"))
        response.raise_for_status()
        return GammaMarket(**response.json())

    def get_events(
            self,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            order: Optional[str] = None,
            ascending: bool = True,
            event_ids: Optional[Union[str, list[str]]] = None,
            slugs: Optional[list[str]] = None,
            archived: Optional[bool] = None,
            active: Optional[bool] = None,
            closed: Optional[bool] = None,
            liquidity_min: Optional[float] = None,
            liquidity_max: Optional[float] = None,
            volume_min: Optional[float] = None,
            volume_max: Optional[float] = None,
            start_date_min: Optional[datetime] = None,
            start_date_max: Optional[datetime] = None,
            end_date_min: Optional[datetime] = None,
            end_date_max: Optional[datetime] = None,
            tag: Optional[str] = None,
            tag_id: Optional[int] = None,
            tag_slug: Optional[str] = None,
            related_tags: bool = False,
    ) -> list[Event]:
        params = {}
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        if order:
            params["order"] = order
            params["ascending"] = ascending
        if event_ids:
            params["id"] = event_ids
        if slugs:
            params["slug"] = slugs
        if archived is not None:
            params["archived"] = archived
        if active is not None:
            params["active"] = active
        if closed is not None:
            params["closed"] = closed
        if liquidity_min:
            params["liquidity_min"] = liquidity_min
        if liquidity_max:
            params["liquidity_max"] = liquidity_max
        if volume_min:
            params["volume_min"] = volume_min
        if volume_max:
            params["volume_max"] = volume_max
        if start_date_min:
            params["start_date_min"] = start_date_min.isoformat()
        if start_date_max:
            params["start_date_max"] = start_date_max.isoformat()
        if end_date_min:
            params["end_date_min"] = end_date_min.isoformat()
        if end_date_max:
            params["end_date_max"] = end_date_max.isoformat()
        if tag:
            params["tag"] = tag
        elif tag_id:
            params["tag_id"] = tag_id
            if related_tags:
                params["related_tags"] = related_tags
        elif tag_slug:
            params["tag_slug"] = tag_slug

        response = self.client.get(self._build_url("/events"), params=params)
        response.raise_for_status()
        return [Event(**event) for event in response.json()]

    def get_event(self, event_id: int) -> Event:
        response = self.client.get(self._build_url(f"/events/{event_id}"))
        response.raise_for_status()
        return Event(**response.json())

    def get_all_events(
            self,
            order: Optional[str] = None,
            ascending: bool = True,
            event_ids: Optional[Union[str, list[str]]] = None,
            slugs: Optional[list[str]] = None,
            archived: Optional[bool] = None,
            active: Optional[bool] = None,
            closed: Optional[bool] = None,
            liquidity_min: Optional[float] = None,
            liquidity_max: Optional[float] = None,
            volume_min: Optional[float] = None,
            volume_max: Optional[float] = None,
            start_date_min: Optional[datetime] = None,
            start_date_max: Optional[datetime] = None,
            end_date_min: Optional[datetime] = None,
            end_date_max: Optional[datetime] = None,
            tag: Optional[str] = None,
            tag_id: Optional[int] = None,
            tag_slug: Optional[str] = None,
            related_tags: bool = False,
    ) -> list[Event]:
        offset = 0
        events = []

        while True:
            part = self.get_events(
                limit=500,
                offset=offset,
                order=order,
                ascending=ascending,
                event_ids=event_ids,
                slugs=slugs,
                archived=archived,
                active=active,
                closed=closed,
                liquidity_min=liquidity_min,
                liquidity_max=liquidity_max,
                volume_min=volume_min,
                volume_max=volume_max,
                start_date_min=start_date_min,
                start_date_max=start_date_max,
                end_date_min=end_date_min,
                end_date_max=end_date_max,
                tag=tag,
                tag_id=tag_id,
                tag_slug=tag_slug,
                related_tags=related_tags,
            )
            events.extend(part)

            if len(part) < 500:
                break

            offset += 500

        return events

    def search_events(self, query: str, active: bool = True) -> list[QueryEvent]:
        """Search for events by query."""
        params = {"q": query, "events_status": "active" if active else "resolved"}
        response = self.client.get("https://polymarket.com/api/events/global", params=params)
        response.raise_for_status()
        return [QueryEvent(**event) for event in response.json()["events"]]

    def grok_market_summary(self, condition_id: str):
        market = self.get_markets(condition_ids=[condition_id])[0]

        params = {
            "marketName": market.group_item_title,
            "eventTitle": market.question,
            "odds": market.outcome_prices[0],
            "marketDescription": market.description,
            "isNegRisk": market.neg_risk,
        }

        with self.client.stream(method="GET", url="https://polymarket.com/api/grok/market-explanation", params=params) as stream:
            messages = []
            citations = []
            for line in stream.iter_lines():
                if line:
                    line_str = line
                    if line_str.startswith("data: "):
                        json_part = line_str[len("data: "):]
                        try:
                            data = json.loads(json_part)
                            # Extract content if present
                            content = data.get("choices", [{}])[0].get("delta", {}).get("content", "")
                            if content:
                                messages.append(content)
                                print(content, end="")  # Stream content
                            # Extract citations if present
                            if "citations" in data:
                                citations.extend(data["citations"])
                        except json.JSONDecodeError:
                            pass

        # After streaming, print citations if any
        if citations:
            print("\n\nCitations:")
            for cite in citations:
                print(f"- {cite}")

    def grok_event_summary(self, event_slug: str):
        params = {
            "prompt": event_slug,
        }

        with self.client.stream(method="GET", url="https://polymarket.com/api/grok/event-summary", params=params) as stream:
            messages = []
            citations = []
            for line in stream.iter_lines():
                if line:
                    line_str = line
                    if line_str.startswith("data: "):
                        json_part = line_str[len("data: "):]
                        try:
                            data = json.loads(json_part)
                            # Extract content if present
                            content = data.get("choices", [{}])[0].get("delta", {}).get("content", "")
                            if content:
                                messages.append(content)
                                print(content, end="")  # Stream content
                            # Extract citations if present
                            if "citations" in data:
                                citations.extend(data["citations"])
                        except json.JSONDecodeError:
                            pass

        # After streaming, print citations if any
        if citations:
            print("\n\nCitations:")
            for cite in citations:
                print(f"- {cite}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.close()



from __future__ import annotations

from datetime import datetime

from pydantic import (
    BaseModel,
    Field,
    Json,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    field_validator,
)

from .common import EthAddress, Keccak256, TimestampWithTZ


class Event(BaseModel):
    # Basic identification
    id: str
    slug: str
    ticker: str | None = None

    # Core event information
    title: str
    description: str | None = None
    resolution_source: str | None = Field(None, alias="resolutionSource")
    category: str | None = None

    # Visual representation
    image: str | None = None
    icon: str | None = None

    # Temporal information
    start_date: datetime | None = Field(None, alias="startDate")
    end_date: datetime | None = Field(None, alias="endDate")
    creation_date: datetime | None = Field(None, alias="creationDate")
    created_at: datetime = Field(alias="createdAt")
    updated_at: datetime | None = Field(None, alias="updatedAt")
    published_at: TimestampWithTZ | None = None
    closed_time: datetime | None = Field(None, alias="closedTime")

    # Status flags
    active: bool
    closed: bool
    archived: bool | None = None
    new: bool | None = None
    featured: bool | None = None
    restricted: bool | None = None
    cyom: bool
    automatically_active: bool | None = Field(None, alias="automaticallyActive")

    # Financial metrics
    liquidity: float | None = None
    volume: float | None = None
    open_interest: int | None = Field(None, alias="openInterest")
    competitive: float | None = None
    volume_24hr: float | None = Field(None, alias="volume24hr")
    liquidity_amm: float | None = Field(None, alias="liquidityAmm")
    liquidity_clob: float | None = Field(None, alias="liquidityClob")

    # Related data
    markets: list[GammaMarket] | None = None
    series: list[Series] | None = None
    tags: list[Tag] | None = None

    # User interaction
    comment_count: int | None = Field(None, alias="commentCount")

    # Display and functionality settings
    sort_by: str | None = Field(None, alias="sortBy")
    show_all_outcomes: bool = Field(alias="showAllOutcomes")
    show_market_images: bool = Field(alias="showMarketImages")
    gmp_chart_mode: str | None = Field(None, alias="gmpChartMode")

    # Negative risk settings
    enable_neg_risk: bool = Field(alias="enableNegRisk")
    neg_risk: bool | None = Field(None, alias="negRisk")
    neg_risk_market_id: str | None = Field(None, alias="negRiskMarketID")
    neg_risk_augmented: bool | None = Field(None, alias="negRiskAugmented")

    # Order book settings
    enable_order_book: bool | None = Field(None, alias="enableOrderBook")


class GammaMarket(BaseModel):
    # Basic identification
    id: str
    slug: str
    condition_id: Keccak256 = Field(alias="conditionId")
    question_id: Keccak256 | None = Field(None, alias="questionID")

    # Core market information
    question: str
    description: str
    resolution_source: str | None = Field(None, alias="resolutionSource")
    outcome: list | None = None
    outcome_prices: Json[list[float]] | list[float] | None = Field(None, alias="outcomePrices")

    # Visual representation
    image: str | None = None
    icon: str | None = None

    # Temporal information
    start_date: datetime | None = Field(None, alias="startDate")
    end_date: datetime | None = Field(None, alias="endDate")
    created_at: datetime = Field(alias="createdAt")
    updated_at: datetime | None = Field(None, alias="updatedAt")
    start_date_iso: datetime | None = Field(None, alias="startDateIso")
    end_date_iso: datetime | None = Field(None, alias="endDateIso")
    deployed_timestamp: datetime | None = Field(None, alias="deployedTimestamp")
    accepting_orders_timestamp: datetime | None = Field(
        None, alias="acceptingOrdersTimestamp",
    )

    # Status flags
    active: bool
    closed: bool
    archived: bool
    new: bool | None = None
    featured: bool | None = None
    restricted: bool
    ready: bool
    deployed: bool | None = None
    funded: bool
    cyom: bool
    approved: bool

    # Financial metrics
    liquidity: float | None = None
    volume: float | None = None
    volume_num: float | None = Field(None, alias="volumeNum")
    liquidity_num: float | None = Field(None, alias="liquidityNum")
    volume_24hr: float | None = Field(None, alias="volume24hr")
    volume_24hr_clob: float | None = Field(None, alias="volume24hrClob")
    volume_clob: float | None = Field(None, alias="volumeClob")
    liquidity_clob: float | None = Field(None, alias="liquidityClob")
    competitive: float | None = None
    spread: float

    # Order book settings
    enable_order_book: bool | None = Field(None, alias="enableOrderBook")
    order_price_min_tick_size: float | None = Field(
        None, alias="orderPriceMinTickSize",
    )
    order_min_size: float | None = Field(None, alias="orderMinSize")
    accepting_orders: bool | None = Field(None, alias="acceptingOrders")

    # Related data
    events: list[Event] | None = None
    clob_rewards: list[ClobReward] | None = Field(None, alias="clobRewards")

    # User interaction
    comment_count: int | None = Field(None, alias="commentCount")

    # Market maker information
    market_maker_address: str = Field(alias="marketMakerAddress")

    # Additional settings
    group_item_title: str | None = Field(None, alias="groupItemTitle")
    group_item_threshold: int | None = Field(None, alias="groupItemThreshold")
    token_ids: Json[list[str]] | list[str] | None = Field(None, alias="clobTokenIds")
    uma_bond: int | None = Field(None, alias="umaBond")
    uma_reward: float | None = Field(None, alias="umaReward")
    neg_risk: bool | None = Field(None, alias="negRisk")
    pager_duty_notification_enabled: bool = Field(alias="pagerDutyNotificationEnabled")
    review_status: str | None = Field(None, alias="reviewStatus")
    rewards_min_size: int = Field(alias="rewardsMinSize")
    rewards_max_spread: float = Field(alias="rewardsMaxSpread")

    # Resolution information
    submitted_by: str | None = None
    resolved_by: EthAddress | None = Field(None, alias="resolvedBy")
    has_reviewed_dates: bool | None = Field(None, alias="hasReviewedDates")

    @field_validator("condition_id", mode="wrap")
    @classmethod
    def validate_condition_id(
            cls, value: str, handler: ValidatorFunctionWrapHandler, info: ValidationInfo,
    ) -> str:
        try:
            # First attempt standard Keccak256 validation
            return handler(value)
        except ValueError:
            active = info.data.get("active", False)

            # Only allow empty string when inactive
            if not active and value == "":
                return value

            # Re-raise original error for other cases
            raise

class ClobReward(BaseModel):
    # Basic identification
    id: str
    condition_id: Keccak256 = Field(alias="conditionId")

    # Reward information
    asset_address: str = Field(alias="assetAddress")
    rewards_amount: float = Field(alias="rewardsAmount")
    rewards_daily_rate: float | None = Field(None, alias="rewardsDailyRate")

    # Temporal information
    start_date: datetime = Field(alias="startDate")
    end_date: datetime = Field(alias="endDate")


class Tag(BaseModel):
    # Basic identification
    id: str
    label: str
    slug: str

    # Display settings
    force_show: bool | None = Field(None, alias="forceShow")
    force_hide: bool | None = Field(None, alias="forceHide")

    # Temporal information
    published_at: TimestampWithTZ | datetime | None = Field(None, alias="publishedAt")
    created_at: datetime | None = Field(None, alias="createdAt")
    updated_at: datetime | None = Field(None, alias="updatedAt")

    # User information
    created_by: int | None = Field(None, alias="createdBy")
    updated_by: int | None = Field(None, alias="updatedBy")


class Series(BaseModel):
    # Basic identification
    id: str
    slug: str
    ticker: str
    title: str

    # Series characteristics
    series_type: str | None = Field(None,  alias="seriesType")
    recurrence: str | None = None
    layout: str | None = None

    # Visual representation
    icon: str | None = None
    image: str | None = None

    # Temporal information
    start_date: datetime | None = Field(None, alias="startDate")
    created_at: datetime = Field(alias="createdAt")
    updated_at: datetime | None = Field(None, alias="updatedAt")
    published_at: TimestampWithTZ | datetime | None = Field(None, alias="publishedAt")

    # Status flags
    active: bool | None = None
    archived: bool | None = None
    closed: bool | None = None
    featured: bool | None = None
    new: bool | None = None
    restricted: bool | None = None

    # Financial metrics
    liquidity: float | None = None
    volume: float | None = None
    volume_24hr: float | None = Field(None, alias="volume24hr")
    competitive: str | None = None

    # User interaction
    comment_count: int = Field(alias="commentCount")
    comments_enabled: bool | None = Field(None, alias="commentsEnabled")

    # User information
    created_by: str | None = Field(None, alias="createdBy")
    updated_by: str | None = Field(None, alias="updatedBy")


class QueryEvent(BaseModel):
    # Basic identification
    id: str
    slug: str
    title: str

    # Visual representation
    image: str | None = None

    # Status flags
    active: bool
    closed: bool
    archived: bool
    neg_risk: bool | None = Field(None, alias="negRisk")

    # Temporal information
    start_date: datetime | None = Field(None, alias="startDate")
    end_date: datetime | None = Field(None, alias="endDate")
    ended: bool

    # Related data
    markets: list[QueryMarket] | None = None


class QueryMarket(BaseModel):
    # Basic identification
    slug: str
    question: str
    group_item_title: str | None = Field(None, alias="groupItemTitle")

    # Market data
    outcomes: list | None = None
    outcome_prices: Json[list[float]] | list[float] | None = Field(None, alias="outcomePrices")
    last_trade_price: float | None = Field(None, alias="lastTradePrice")
    best_ask: float | None = Field(None, alias="bestAsk")
    best_bid: float | None = Field(None, alias="bestBid")
    spread: float

    # Status flags
    closed: bool
    archived: bool
